(**************************************************************************)
(* Streamex: sample code to show Streams and Okasaki's Queues             *)
(*                                                                        *)
(*  Author(s):  Roberto Di Cosmo                                          *)
(*                                                                        *)
(*  This library is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU Lesser General Public License as        *)
(*  published by the Free Software Foundation, either version 2 of the    *)
(*  License, or (at your option) any later version.  A special linking    *)
(*  exception to the GNU Lesser General Public License applies to this    *)
(*  library, see the LICENSE file for more information.                   *)
(**************************************************************************)

(* streams suspendus des le debut *)

type 'a stream = Nil | Cons of 'a * 'a strsusp
and 'a strsusp = 'a stream Lazy.t;;

(* test avec le factoriel *)

let rec fact_from n = 
    lazy (Cons((Printf.eprintf "*%!";fact n), 
            lazy (Lazy.force (fact_from (n+1)))));;

let rec take_list n s = match (n,s) with
      0,s -> [] 
    | n,s -> match (Lazy.force s) with Nil -> [] 
           | Cons(v,r) -> v :: (take_list (n-1) r);;

let fact_nat = fact_from 0;;
take_list 5 fact_nat;;
take_list 5 fact_nat;;


(* le module Stream et la signature STREAM *)

module type STREAM = sig
  type 'a streamhead = Nil | Cons of 'a * 'a stream
  and 'a stream = 'a streamhead Lazy.t

  (* concatenation de streams *)
  val (++) : 'a stream -> 'a stream -> 'a stream

  (* un stream contenant les premiers n elements *)
  val take : int -> 'a stream -> 'a stream

  (* le stream sans les premiers n elements *)
  val drop : int -> 'a stream -> 'a stream
  val reverse : 'a stream -> 'a stream
end;;

let (!$) = Lazy.force;; (* abbreviation *)

module Stream : STREAM = struct
  type 'a streamhead = Nil | Cons of 'a * 'a stream
  and 'a stream = 'a streamhead Lazy.t

  (* concatenation *)
  let (++) s1 s2 = 
    let rec aux s = match !$s with
    | Nil -> !$s2
    | Cons (hd, tl) -> Cons (hd, lazy(aux tl))
  in lazy(aux s1)

  (* copie paresseuse des premier n elements *)
  let take n s = 
    let rec take' n s = match n, !$s with
    | 0, _ -> Nil
    | _, Nil -> Nil
    | _, Cons (hd, tl) -> Cons (hd, lazy (take' (n - 1) tl))
  in lazy(take' n s)

  (* copie du stream apres avoir n elements *)
  let drop n s =
    let rec drop' n s = 
      match n with 0 -> !$s
      | n -> match !$s with
             | Nil -> Nil
             | Cons (_, tl) -> (drop' (n - 1) tl) in
    match n with 0 -> Nil | n -> lazy (drop' n s)

  (* retourne un stream, monolythique...! *)
  let reverse s =
    let rec reverse' acc s = match !$s with
      | Nil -> acc
      | Cons (hd, tl) -> 
           reverse' (Cons (hd, lazy acc)) tl in
    lazy (reverse' Nil s)

  let add x s = lazy (Cons(x,s))  
end;;

open Stream;;

(* Okasaki : files avec streams *)

module type FIFO = sig
  type 'a t
  exception Empty
  val empty : 'a t
  val is_empty : 'a t -> bool
  val add : 'a t -> 'a -> 'a t  
  val remove : 'a t -> 'a * 'a t
  (** leve l'exception [Empty] sur une file vide *)
end;;

module FileDS : FIFO = struct
  type 'a t = int * 'a stream * int * 'a stream
  exception Empty

  let empty = 0, lazy Nil, 0, lazy Nil
  let is_empty (lenf, _, _, _) = lenf = 0

  let check (lenf, f, lenr, r as q) =
    if lenr <= lenf then q
    else (lenf + lenr, f ++ reverse r, 0, lazy Nil)

  let add (lenf, f, lenr, r) x = check (lenf, f, lenr + 1, lazy(Cons (x, r)))

  let remove (lenf,f,lenr,r) = match Lazy.force f with
    | Nil          -> raise Empty
    | Cons (x, f') -> x,check (lenf - 1, f', lenr, r)
end;;

(* version instrumentee des streams et de File DS pour tracer l'execution *)

module Stream : STREAM = struct
  type 'a streamhead = Nil | Cons of 'a * 'a stream
  and 'a stream = 'a streamhead Lazy.t

  (* concatenation *)
  let (++) s1 s2 = 
    let rec aux s = match !$s with
    | Nil -> !$s2
    | Cons (hd, tl) -> Cons (hd, lazy(aux tl))
  in lazy(aux s1)

  (* copie paresseuse des premier n elements *)
  let take n s = 
    let rec take' n s = match n, !$s with
    | 0, _ -> Nil
    | _, Nil -> Nil
    | _, Cons (hd, tl) -> Cons (hd, lazy (take' (n - 1) tl))
  in lazy(take' n s)

  (* copie du stream apres avoir n elements *)
  let drop n s =
    let rec drop' n s = 
      match n with 0 -> !$s
      | n -> match !$s with
             | Nil -> Nil
             | Cons (_, tl) -> (drop' (n - 1) tl) in
    lazy (drop' n s)

  let rec str_len = function Nil -> 0 | Cons(_,tl) -> 1+(str_len !$tl)

  (* retourne un stream, monolythique...! *)
  let reverse s =
    let rec reverse' acc s = match !$s with
      | Nil -> acc
      | Cons (hd, tl) -> 
           reverse' (Cons (hd, lazy acc)) tl in
    lazy (let s' = reverse' Nil s in
          Printf.eprintf "Reverse stream: cout %d\n%!" (str_len s'); s')

  let add x s = lazy (Cons(x,s))  

end;;

module FileDS : FIFO = struct
  type 'a t = int * 'a stream * int * 'a stream
  exception Empty

  let empty = 0, lazy Nil, 0, lazy Nil
  let is_empty (lenf, _, _, _) = lenf = 0

  let check (lenf, f, lenr, r as q) =
    if lenr <= lenf then q
    else begin
      Printf.eprintf "Ajoutee suspension de taille %d; front = %d\n%!" lenr (lenr+lenf);
      (lenf + lenr, f ++ reverse r, 0, lazy Nil)
    end

  let add (lenf, f, lenr, r) x = check (lenf, f, lenr + 1, lazy(Cons (x, r)))

  let remove (lenf,f,lenr,r) = match Lazy.force f with
    | Nil          -> raise Empty
    | Cons (x, f') -> x,check (lenf - 1, f', lenr, r)
end;;

(*********** tests de l'efficacite de la solution ******)

let ext_intv startv endv =
  let s,e = (min startv endv),(max startv endv) in
  let rec aux acc = function n -> if n=s then n::acc else aux (n::acc) (n-1)
  in aux [] e

let enfile n s = List.fold_left FileDS.add s (ext_intv 1 n);;

let to_list f = 
   let rec aux acc f = 
     try 
       let x,f' = FileDS.remove f in x::(aux acc f')
     with _ -> acc
   in aux [] f;;

let e0 = FileDS.empty;;

(* e14 est dans une config comme dans le cours *)
let e14 = enfile 14 e0;;

(* le premier remove ajoute un reverse de longueur 7 *)
let e13 = snd(FileDS.remove e14);;
let e12 = snd(FileDS.remove e13);;
let e11 = snd(FileDS.remove e12);;
let e10 = snd(FileDS.remove e11);;
let e9 = snd(FileDS.remove e10);;
let e8 = snd(FileDS.remove e9);;
let e7 = snd(FileDS.remove e8);;

(* ceci declenche le reverse ajoute par e13 *)

let e6 = snd(FileDS.remove e7);;

(* et maintenant, voyons que relancer e7, e6 ne recalcule pas *)

let e7' = snd(FileDS.remove e8);;
let e6' = snd(FileDS.remove e7');;


(* quelques copies a la racine *)
let e6_1= e6;;
let e6_2= e6;;
let e6_3= e6;;



(* files avec streams, version fonctorielle *)

module FileDS (S: STREAM): FIFO = struct
  type 'a t = int * 'a S.stream * int * 'a S.stream
  exception Empty

  let empty = 0, lazy S.Nil, 0, lazy S.Nil
  let is_empty (lenf, _, _, _) = lenf = 0

  let check (lenf, f, lenr, r as q) =
    if lenr <= lenf then q
    else (lenf + lenr, S.(++) f (S.reverse r), 0, lazy S.Nil)

  let add (lenf, f, lenr, r) x = check (lenf, f, lenr + 1, lazy(S.Cons (x, r)))

  let remove (lenf,f,lenr,r) = match Lazy.force f with
    | S.Nil          -> raise Empty
    | S.Cons (x, f') -> x,check (lenf - 1, f', lenr, r)
end;;



(***********************************************************************)
(*                   Files Temps Reel de Okasaki                       *)
(***********************************************************************)
(******************************************************************************)
(*                          Version Roberto notes de cours                    *)
(******************************************************************************)

module RealTimeQueue : FIFO = struct
type 'a t = 'a stream * 'a list * 'a stream
exception Empty

let empty = lazy Nil, [], lazy Nil

let is_empty (f, _, _) = 
    match !$f with Nil -> true | _ -> false

let rec rotate (f,r,a) = Printf.eprintf "rot\n%!"; (*trace*)
  match (!$f,r,a) with
  | Nil, [y], a -> lazy (Cons (y, a))
  | Cons (x, xs), y :: ys, a ->
      lazy (Cons (x, 
                 (rotate (xs, ys, lazy (Cons (y, a))))))
  | _, _, _ -> failwith "rotate: configuration illegale"

let exec (f,r,s) = match (f,r,!$s) with
  | f, r, Cons (x, s) -> f, r, s
  | f, r, Nil -> let f' = rotate (f, r, lazy Nil)
                 in f', [], f'

let add (f, r, s) x = exec (f, x :: r, s)

let remove (f,r,s) = match (!$f,r,s) with
  | Nil, _, _ -> raise Empty
  | Cons (x, f), r, s -> x, exec (f, r, s)
end;;

let enfilert n s = 
  List.fold_left 
    (fun s -> fun n -> Printf.eprintf "add: %d\n%!" n;RealTimeQueue.add s n)
    s
    (ext_intv 1 n);;

let rec depile = function
    0,_ -> []
  | n,q -> Printf.eprintf "-\n%!";let x,q' = RealTimeQueue.remove q in x::(depile ((n-1),q'));;


let rec list_to_stream = function
  | [] -> lazy Nil
  | x :: xs -> lazy (Cons (x, list_to_stream xs))

let rec stream_to_list s = match !$s with
  | Nil -> []
  | Cons(x,xs) -> Printf.eprintf "+\n%!"; x::(stream_to_list xs)

(* on teste et on observe que rotate n'est pas appele imbrique *)

let q7=enfilert 7 RealTimeQueue.empty;;
let q0 = depile (7,q7);;


(******************************************************************************)
(*                   version avec l'interface de Okasaki                      *)
(******************************************************************************)

exception Empty
exception Impossible_pattern of string

let impossible_pat x = raise (Impossible_pattern x)

module type QUEUE = sig
  type 'a queue

  val empty : 'a queue
  val is_empty : 'a queue -> bool

  val snoc : 'a queue -> 'a -> 'a queue
  val head : 'a queue -> 'a        (* raises Empty if queue is empty *)
  val tail : 'a queue -> 'a queue  (* raises Empty if queue is empty *)
end

(* adapted to fully lazy streams *)

module RealTimeQueue (* : QUEUE *) = struct
  type 'a queue = 'a stream * 'a list * 'a stream

  let empty = lazy Nil, [], lazy Nil

  let is_empty (f, _, _) = match !$f with Nil -> true | _ -> false

  let rec rotate (f,s,r) = match (!$f,s,r) with
    | Nil, y :: _, a -> lazy (Cons (y, a))
    | Cons (x, xs), y :: ys, a ->
        lazy (Cons (x, (rotate (xs, ys, lazy (Cons (y, a))))))
    | _, [], _ -> impossible_pat "rotate"

  let exec (f,r,s) = match (f,r,!$s) with
    | f, r, Cons (x, s) -> f, r, s
    | f, r, Nil -> let f' = rotate (f, r, lazy Nil) in f', [], f'

  let snoc (f, r, s) x = exec (f, x :: r, s)

  let head (f, _, _) = match !$f with
    | Nil -> raise Empty
    | Cons (x, _) -> x

  let tail (f,r,s) = match (!$f,r,s) with
    | Nil, _, _ -> raise Empty
    | Cons (_, f), r, s -> exec (f, r, s)
end


(******************************************************************************)
(* Code pour les files qui marche mal en mode persistant *)

module FifoDLCount = struct
  exception Empty
  type 'a t = 'a list * 'a list
  let empty = ([], [])
  let is_empty = 
  function
    ([],[]) -> true 
  |   _     -> false

  (* on ajoute de compteurs pour tracer le cout *)
  let ncalls = ref 0
  let cost = ref 0
  let incr c n = c:= !c +n
  let reset c = c := 0

  let add x (l1, l2) = 
      incr ncalls 1; incr cost 1; (x::l1, l2)

  let remove (l1, l2) = incr ncalls 1;
  match l2 with
  | a::l -> incr cost 1; (a, (l1, l))
  | [] -> incr cost (List.length l1);
      match List.rev l1 with
      | [] -> raise Empty
      | a::l -> (a, ([], l))
end;;
open FifoDLCount
let e = empty;;
let l = List.fold_left (fun l x -> add x l)
        e [1;2;3;4;5];;
ncalls;;
cost;;
(* Mais le cout n'est plus lineaire si on utilise
   la structure de facon persistante !!! *)

let (v1,l1) = remove l;;
let (v2,l2) = remove l;;
let (v3,l3) = remove l;;
let (v4,l4) = remove l;;
let (v5,l5) = remove l;;
(* let (v6,l6) = remove l;; *)

ncalls;;
cost;;
